"""Generate a status line for swaybar."""

import argparse
import logging
import tomllib
from pathlib import Path
from typing import Iterable

from .config import config
from .element import BaseElement
from .env import config_home, environ_path, environ_paths, self_name
from .logging import logger
from .loop import start
from .modules import Modules


def configure_logging(level: str) -> None:
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(logging.Formatter("%(name)s: %(levelname)s: %(message)s"))
    logging.basicConfig(level=level.upper(), handlers=[stream_handler])


def parse_args() -> argparse.Namespace:
    class MyHelpFormatter(argparse.RawDescriptionHelpFormatter):
        def __init__(self, prog) -> None:
            super().__init__(prog, indent_increment=4, max_help_position=30)

    parser = argparse.ArgumentParser(description=__doc__, formatter_class=MyHelpFormatter)
    parser.add_argument(
        "-c",
        "--config-file",
        metavar="FILE",
        type=Path,
        help="override configuration file",
    )
    parser.add_argument(
        "-C",
        "--config-dir",
        metavar="DIRECTORY",
        type=Path,
        help="override configuration directory",
    )
    parser.add_argument(
        "-I",
        "--include",
        action="append",
        metavar="DIRECTORY",
        type=Path,
        help="include additional modules package",
    )
    parser.add_argument(
        "-i",
        "--interval",
        type=float,
        metavar="SECONDS",
        help="override default update interval",
    )
    parser.add_argument(
        "--no-click-events",
        dest="click_events",
        action="store_false",
        help="disable click events",
    )
    parser.add_argument(
        "--log-level",
        metavar="LEVEL",
        default="info",
        choices=["debug", "info", "warning", "error", "critical"],
        help="override default minimum logging level (default: %(default)s)",
    )
    return parser.parse_args()


def parse_config(args):
    config_dir = args.config_dir or environ_path("SWAYSTATUS_CONFIG_DIR") or (config_home / self_name)
    config_file = args.config_file or environ_path("SWAYSTATUS_CONFIG_FILE") or (config_dir / "config.toml")

    if config_file.is_file():
        config.update(tomllib.loads(open(config_file).read()))

    config["include"] = (
        (args.include or [])
        + [config_dir / "modules"]
        + [Path(d).expanduser() for d in config.get("include", [])]
        + (environ_paths("SWAYSTATUS_MODULE_PATH") or [])
    )

    if args.interval:
        config["interval"] = args.interval

    if not args.click_events:
        config["click_events"] = False

    return config


def deep_merge_dicts(first: dict, second: dict) -> dict:
    """Recursively merge the second dictionary into the first."""
    result = first.copy()
    for key, value in second.items():
        if key in result and isinstance(result[key], dict) and isinstance(value, dict):
            result[key] = deep_merge_dicts(result[key], value)
        else:
            result[key] = value
    return result


def load_elements(order: Iterable[str], include: Iterable[Path], settings: dict) -> list[BaseElement]:
    elements = []
    modules = Modules(include)

    for key in order:
        try:
            name, instance = key.split(":", maxsplit=1)
        except ValueError:
            name, instance = key, None

        module = modules.find(name)
        logger.info(f"Loaded module from file: {module.__file__}")

        kwargs = deep_merge_dicts(settings.get(name, {}), settings.get(key, {}))
        kwargs.update({"name": name, "instance": instance})

        logger.debug(f"Initializing module: {kwargs!r}")
        elements.append(module.Element(**kwargs))

    return elements


def main() -> int:
    args = parse_args()
    configure_logging(args.log_level)

    config = parse_config(args)
    logger.debug(f"Using configuration: {config!r}")

    elements = load_elements(
        config["order"],
        config["include"],
        config["settings"],
    )

    try:
        start(
            elements,
            config["interval"],
            config["click_events"],
        )
    except Exception:
        logger.exception("Unhandled exception in main loop")
        return 1
    return 0



import sys
from functools import cached_property
from importlib import import_module, metadata
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from types import ModuleType
from typing import Iterable
from uuid import uuid4


def unique_package_name() -> str:
    return str(uuid4()).replace("-", "")


class Modules:
    def __init__(self, include: Iterable[Path]):
        self.include = list(include)
        self.cache: dict[str, ModuleType] = {}

    @cached_property
    def packages(self) -> list[str]:
        result = []

        for i, modules_dir in enumerate(self.include):
            if (init_file := modules_dir.expanduser() / "__init__.py").is_file():
                package_name = unique_package_name()
                if spec := spec_from_file_location(package_name, init_file):
                    package = module_from_spec(spec)
                    sys.modules[package_name] = package
                    if spec.loader:
                        spec.loader.exec_module(package)
                        result.append(package_name)

        for entry_point in metadata.entry_points(group="swaystatus.modules"):
            result.append(entry_point.load().__name__)

        return result

    def find(self, name: str) -> ModuleType:
        if name not in self.cache:
            for package in self.packages:
                try:
                    self.cache[name] = import_module(f"{package}.{name}")
                    break
                except ModuleNotFoundError:
                    continue
            else:
                raise ModuleNotFoundError(f"Module not found in any package: {name}")

        return self.cache[name]



from pathlib import Path

from swaystatus.element import BaseElement


def test_element_udpate_no_output():
    """Ensure that nothing is added to the output by default."""
    output = []
    BaseElement().on_update(output)
    assert len(output) == 0


def test_element_on_click_method():
    """Ensure that click event handlers can be defined as a method."""
    hit = False

    class Element(BaseElement):
        def on_click_1(self, event):
            nonlocal hit
            hit = True

    Element().on_click({"button": 1})
    assert hit


def test_element_on_click_function():
    """Ensure that function click event handlers can be set at initialization."""
    hit = False

    def handler(event):
        nonlocal hit
        hit = True

    BaseElement(on_click={1: handler}).on_click({"button": 1})
    assert hit


def test_element_on_click_shell(tmp_path):
    """Ensure that shell command click event handlers can be set at initialization."""
    button = 1
    cases = {
        "${foo}": "some string",  # environment variables added
        "${button}": str(button),  # environment variables from event
        "~": str(Path.home()),  # shell tilde expansion
    }
    env = {"foo": cases["${foo}"]}
    event = {"button": button}
    tmp_path.mkdir(parents=True, exist_ok=True)
    stdout_file = tmp_path / "stdout"

    for s, expected in cases.items():
        handler = f"echo {s} >{stdout_file}"  # shell redirection
        BaseElement(on_click={1: handler}, env=env).on_click(event).wait()
        assert stdout_file.read_text().strip() == expected


def test_element_create_block_default():
    """Ensure that when no name or instance is set, it's not included in the output."""
    assert BaseElement().create_block("test") == {"full_text": "test"}


def test_element_create_block_with_id_set_at_init():
    """Ensure that name and instance can be set at initialization."""
    element = BaseElement(name="foo", instance="bar")
    assert element.create_block("test") == {
        "full_text": "test",
        "name": element.name,
        "instance": element.instance,
    }


def test_element_create_block_with_id_set_after_init():
    """Ensure that name and instance can be overridden after initialization."""
    element = BaseElement()
    element.name = "foo"
    element.instance = "bar"
    assert element.create_block("test") == {
        "full_text": "test",
        "name": element.name,
        "instance": element.instance,
    }


def test_element_create_block_with_id_set_in_block():
    """Ensure that name and instance can be overridden per-block."""
    element = BaseElement(name="foo", instance="bar")
    assert element.create_block("test", name="baz", instance="qux") == {
        "full_text": "test",
        "name": "baz",
        "instance": "qux",
    }


def test_element_create_block_with_kwargs():
    """Ensure that keyword arguments passed to `create_block` are included in the result."""
    kwargs = {"foo": "a", "bar": "b"}
    assert BaseElement().create_block("test", **kwargs) == dict(full_text="test", **kwargs)



from typing import Any

from .element import BaseElement

type Output = list[dict[str, Any]]


class Updater:
    def __init__(self, *elements: BaseElement) -> None:
        super().__init__()
        self.elements = list(elements)

    def update(self) -> Output:
        """
        Prompt every element for any updates to the status bar.

        It does this by giving each element a turn at appending blocks to an
        `output` list that it passes as the first argument to the element's
        `on_update` method. This is done in the order that the elements were
        given to the updater at initialization.
        """
        output: Output = []
        for element in self.elements:
            element.on_update(output)
        return output



import json
import locale
import sys
from functools import cached_property
from signal import SIGCONT, SIGINT, SIGSTOP, SIGTERM, SIGUSR1, Signals, signal
from threading import Event, Thread
from typing import Iterable, Iterator

from .element import BaseElement
from .logging import logger
from .updater import Updater

type Seconds = float

header = {
    "version": 1,
    "stop_signal": SIGSTOP,
    "cont_signal": SIGCONT,
    "click_events": False,
}
body_start = "[[]"
body_item = ",{}"


class OutputWriter:
    file = sys.stdout

    def __init__(self, updater: Updater, interval: Seconds, click_events: bool) -> None:
        self.updater = updater
        self.interval = interval
        self.click_events = click_events
        self._tick = Event()
        self._running = Event()

    def send(self, line: str) -> None:
        print(line, flush=True, file=self.file)

    def update(self) -> None:
        self._tick.set()

    def stop(self) -> None:
        self._running.clear()
        self._tick.set()

    def start(self) -> None:
        self.send(json.dumps(header | dict(click_events=self.click_events)))
        self.send(body_start)
        self._running.set()
        while self._running.is_set():
            self.send(body_item.format(json.dumps(self.updater.update())))
            self._tick.clear()
            self._tick.wait(self.interval)


class InputReader(Thread):
    daemon = True
    file = sys.stdin

    def __init__(self, elements: Iterable[BaseElement], output_writer: OutputWriter) -> None:
        super().__init__(name="input")
        self.elements = elements
        self.output_writer = output_writer

    @cached_property
    def elements_by_key(self) -> dict[str, BaseElement]:
        return {key: elem for elem in self.elements if (key := elem.key())}

    @property
    def click_events(self) -> Iterator[dict]:
        assert self.file.readline().strip() == "["
        for line in self.file:
            yield json.loads(line.strip().lstrip(","))

    def run(self) -> None:
        logger.info("Listening for click events from stdin...")

        for click_event in self.click_events:
            logger.debug(f"Received click event: {click_event!r}")

            name = click_event["name"]
            instance = click_event.get("instance")
            key = f"{name}:{instance}" if instance else name

            try:
                element = self.elements_by_key[key]
            except KeyError:
                element = self.elements_by_key[name]

            element.on_click(click_event)
            self.output_writer.update()


def start(elements: Iterable[BaseElement], interval: Seconds, click_events: bool) -> None:
    locale.setlocale(locale.LC_ALL, "")

    updater = Updater(*elements)

    output_writer = OutputWriter(updater, interval, click_events)

    if click_events:
        input_thread = InputReader(elements, output_writer)

    def update(sig, frame):
        logger.info(f"Signal was sent to update: {Signals(sig).name} ({sig})")
        logger.debug(f"Current stack frame: {frame}")
        output_writer.update()

    signal(SIGUSR1, update)

    def shutdown(sig, frame):
        logger.info(f"Signal was sent to shutdown: {Signals(sig).name} ({sig})")
        logger.debug(f"Current stack frame: {frame}")
        output_writer.stop()

    signal(SIGINT, shutdown)
    signal(SIGTERM, shutdown)

    if click_events:
        input_thread.start()

    output_writer.start()



from subprocess import PIPE, Popen
from threading import Thread
from typing import IO, Callable

type StreamHandler = Callable[[str], None]


class ProxyThread(Thread):
    def __init__(self, source: IO[str], handler: StreamHandler) -> None:
        super().__init__()
        self.source = source
        self.handler = handler

    def run(self) -> None:
        with self.source as lines:
            for line in lines:
                self.handler(line)


class PopenStreamHandler(Popen):
    """Just like `Popen`, but handle stdout and stderr output in dedicated threads."""

    def __init__(self, stdout_handler, stderr_handler, *args, **kwargs) -> None:
        kwargs["stdout"] = kwargs["stderr"] = PIPE
        super().__init__(*args, **kwargs)
        assert self.stdout and self.stderr
        ProxyThread(self.stdout, stdout_handler).start()
        ProxyThread(self.stderr, stderr_handler).start()



import os
import sys
from pathlib import Path

self_name = os.path.basename(sys.argv[0])

cache_home = Path(os.environ.get("XDG_CACHE_HOME", "~/.cache")).expanduser()
config_home = Path(os.environ.get("XDG_CONFIG_HOME", "~/.config")).expanduser()
data_home = Path(os.environ.get("XDG_DATA_HOME", "~/.local/share")).expanduser()
state_home = Path(os.environ.get("XDG_STATE_HOME", "~/.local/state")).expanduser()


def environ_path(name: str) -> Path | None:
    if value := os.environ.get(name):
        return Path(value).expanduser()
    return None


def environ_paths(name: str) -> list[Path] | None:
    if value := os.environ.get(name):
        return [Path(p).expanduser() for p in value.split(":")]
    return None



import shutil
from pathlib import Path

import pytest

from swaystatus import modules


def copy_module(name, directory: Path):
    """Copy a test module to a package directory."""
    src = Path(__file__).parent / "modules" / f"{name}.py"
    dst = directory / src.name

    directory.mkdir(parents=True, exist_ok=True)
    (directory / "__init__.py").touch()
    shutil.copyfile(src, dst)

    return dst


def test_modules_find_module_not_found():
    """Ensure that requesting a non-existent module will raise an error."""
    with pytest.raises(ModuleNotFoundError, match="foo"):
        modules.Modules([]).find("foo")


def test_modules_find(tmp_path):
    """Ensure that an existing module will be found in a valid package."""
    path = copy_module("no_output", tmp_path)
    assert modules.Modules([tmp_path]).find("no_output").__file__ == str(path)


def test_modules_entry_points(tmp_path, monkeypatch):
    """Ensure that module packages defined as an entry point are recognized."""

    class Package:
        __name__ = "test"

    class EntryPoint:
        def load(self):
            return Package()

    def entry_points(**kwargs):
        assert kwargs["group"] == "swaystatus.modules"
        return [EntryPoint()]

    monkeypatch.setattr(modules.metadata, "entry_points", entry_points)
    copy_module("no_output", tmp_path)

    packages = modules.Modules([tmp_path]).packages
    assert len(packages) == 2
    assert packages[-1] == "test"



from swaystatus.element import BaseElement
from swaystatus.updater import Updater


def test_updater_no_blocks():
    """Ensure that if an element does not emit any blocks, none appear in the output."""

    class NoBlocks(BaseElement):
        def on_update(self, output):
            pass

    updater = Updater(NoBlocks())

    assert updater.update() == []


def test_updater_multiple_blocks():
    """Ensure that a single element is able to output multiple blocks."""

    texts = ["foo", "bar", "baz"]

    class MultipleBlocks(BaseElement):
        def on_update(self, output):
            output.extend([self.create_block(text) for text in texts])

    updater = Updater(MultipleBlocks())

    assert updater.update() == [dict(full_text=text) for text in texts]


def test_updater_multiple_elements():
    """Ensure that multiple elements output their blocks in the correct order."""

    class Foo(BaseElement):
        def on_update(self, output):
            output.append(self.create_block("foo"))

    class Bar(BaseElement):
        def on_update(self, output):
            output.append(self.create_block("bar"))

    updater = Updater(Foo(), Bar())

    assert updater.update() == [dict(full_text="foo"), dict(full_text="bar")]
